@@ -19,6 +19,7 @@
 #endif
 #include <datatype/timestamp.h>
 #include <common/pg_lzcompress.h>
+#include <inttypes.h>
 #include <string.h>
 #include <ctype.h>
 #include <stdio.h>
@@ -430,7 +431,7 @@ CopyAppendNumeric(const char *buffer, int num_size)
 		}
 		else
 		{
-			ndigits = num_size / sizeof(NumericDigit);
+			ndigits = (num_size - NUMERIC_HEADER_SIZE(num)) / sizeof(NumericDigit);
 			digits = (NumericDigit *) ((char *) num + NUMERIC_HEADER_SIZE(num));
 			i = (weight + 1) * DEC_DIGITS;
 			if (i <= 0)
@@ -724,7 +725,7 @@ decode_uint(const char *buffer, unsigned int buff_size, unsigned int *out_size)
 static int
 decode_bigint(const char *buffer, unsigned int buff_size, unsigned int *out_size)
 {
-	const char *new_buffer = (const char *) LONGALIGN(buffer);
+	const char *new_buffer = (const char *) DOUBLEALIGN(buffer);
 	unsigned int delta = (unsigned int) ((uintptr_t) new_buffer - (uintptr_t) buffer);
 
 	if (buff_size < delta)
@@ -745,7 +746,7 @@ decode_bigint(const char *buffer, unsigned int buff_size, unsigned int *out_size
 static int
 decode_time(const char *buffer, unsigned int buff_size, unsigned int *out_size)
 {
-	const char *new_buffer = (const char *) LONGALIGN(buffer);
+	const char *new_buffer = (const char *) DOUBLEALIGN(buffer);
 	unsigned int delta = (unsigned int) ((uintptr_t) new_buffer - (uintptr_t) buffer);
 	int64		timestamp,
 				timestamp_sec;
@@ -763,7 +764,7 @@ decode_time(const char *buffer, unsigned int buff_size, unsigned int *out_size)
 	timestamp_sec = timestamp / 1000000;
 	*out_size = sizeof(int64) + delta;
 
-	CopyAppendFmt("%02" INT64_MODIFIER "d:%02" INT64_MODIFIER "d:%02" INT64_MODIFIER "d.%06" INT64_MODIFIER "d",
+	CopyAppendFmt("%02" PRIu64 ":%02" PRIu64 ":%02" PRIu64 ".%06" PRIu64,
 				  timestamp_sec / 60 / 60, (timestamp_sec / 60) % 60, timestamp_sec % 60,
 				  timestamp % 1000000);
 
@@ -774,7 +775,7 @@ decode_time(const char *buffer, unsigned int buff_size, unsigned int *out_size)
 static int
 decode_timetz(const char *buffer, unsigned int buff_size, unsigned int *out_size)
 {
-	const char *new_buffer = (const char *) LONGALIGN(buffer);
+	const char *new_buffer = (const char *) DOUBLEALIGN(buffer);
 	unsigned int delta = (unsigned int) ((uintptr_t) new_buffer - (uintptr_t) buffer);
 	int64		timestamp,
 				timestamp_sec;
@@ -796,7 +797,7 @@ decode_timetz(const char *buffer, unsigned int buff_size, unsigned int *out_size
 	tz_min = -(tz_sec / 60);
 	*out_size = sizeof(int64) + sizeof(int32) + delta;
 
-	CopyAppendFmt("%02" INT64_MODIFIER "d:%02" INT64_MODIFIER "d:%02" INT64_MODIFIER "d.%06" INT64_MODIFIER "d%c%02d:%02d",
+	CopyAppendFmt("%02" PRIu64 ":%02" PRIu64 ":%02" PRIu64 ".%06" PRIu64 "%c%02d:%02d",
 				  timestamp_sec / 60 / 60, (timestamp_sec / 60) % 60, timestamp_sec % 60,
 				  timestamp % 1000000, (tz_min > 0 ? '+' : '-'), abs(tz_min / 60), abs(tz_min % 60));
 
@@ -850,7 +851,7 @@ decode_date(const char *buffer, unsigned int buff_size, unsigned int *out_size)
 static int
 decode_timestamp_internal(const char *buffer, unsigned int buff_size, unsigned int *out_size, bool with_timezone)
 {
-	const char *new_buffer = (const char *) LONGALIGN(buffer);
+	const char *new_buffer = (const char *) DOUBLEALIGN(buffer);
 	unsigned int delta = (unsigned int) ((uintptr_t) new_buffer - (uintptr_t) buffer);
 	int64		timestamp,
 				timestamp_sec;
@@ -898,7 +899,7 @@ decode_timestamp_internal(const char *buffer, unsigned int buff_size, unsigned i
 	j2date(jd, &year, &month, &day);
 	timestamp_sec = timestamp / 1000000;
 
-	CopyAppendFmt("%04d-%02d-%02d %02" INT64_MODIFIER "d:%02" INT64_MODIFIER "d:%02" INT64_MODIFIER "d.%06" INT64_MODIFIER "d%s%s",
+	CopyAppendFmt("%04d-%02d-%02d %02" PRIu64 ":%02" PRIu64 ":%02" PRIu64 ".%06" PRIu64 "%s%s",
 				  (year <= 0) ? -year + 1 : year, month, day,
 				  timestamp_sec / 60 / 60, (timestamp_sec / 60) % 60, timestamp_sec % 60,
 				  timestamp % 1000000,
@@ -1077,17 +1078,6 @@ extract_data(const char *buffer, unsigned int buff_size, unsigned int *out_size,
 	int			padding = 0;
 	int			result	= 0;
 
-	/* Skip padding bytes. */
-	while (*buffer == 0x00)
-	{
-		if (buff_size == 0)
-			return -1;
-
-		buff_size--;
-		buffer++;
-		padding++;
-	}
-
 	if (VARATT_IS_1B_E(buffer))
 	{
 		/*
@@ -1152,6 +1142,11 @@ extract_data(const char *buffer, unsigned int buff_size, unsigned int *out_size,
 		return result;
 	}
 
+	/* Skip padding bytes. */
+	padding = (char *)INTALIGN(buffer) - buffer;
+	buffer += padding;
+	buff_size -= padding;
+
 	if (VARATT_IS_4B_U(buffer) && buff_size >= 4)
 	{
 		/*
@@ -1363,16 +1358,21 @@ ReadStringFromToast(const char *buffer,
 		varatt_external toast_ptr;
 		char	   *toast_data = NULL;
 		/* Number of chunks the TOAST data is divided into */
-		int32		num_chunks;
+		uint32		num_chunks;
+		/* Next chunk to read */
+		uint32		want_chunk_id = 0;
 		/* Actual size of external TOASTed value */
 		int32		toast_ext_size;
-		/* Path to directory with TOAST realtion file */
+		/* Path to directory with TOAST relation file */
 		char	   *toast_relation_path;
 		/* Filename of TOAST relation file */
 		char		toast_relation_filename[MAXPGPATH];
 		FILE	   *toast_rel_fp;
+		unsigned int toast_relation_block_size;
+		unsigned int toastDataRead = 0;
 		unsigned int block_options = 0;
 		unsigned int control_options = 0;
+		int			loops = 0;
 
 		VARATT_EXTERNAL_GET_POINTER(toast_ptr, buffer);
 
@@ -1382,7 +1382,7 @@ ReadStringFromToast(const char *buffer,
 #else
 		toast_ext_size = toast_ptr.va_extsize;
 #endif
-		num_chunks = (toast_ext_size - 1) / TOAST_MAX_CHUNK_SIZE + 1;
+		num_chunks = ((toast_ext_size - 1) / TOAST_MAX_CHUNK_SIZE) + 1;
 
 		printf("  TOAST value. Raw size: %8d, external size: %8d, "
 				"value id: %6d, toast relation id: %6d, chunks: %6d\n",
@@ -1398,46 +1398,63 @@ ReadStringFromToast(const char *buffer,
 		sprintf(toast_relation_filename, "%s/%d",
 				*toast_relation_path ? toast_relation_path : ".",
 				toast_ptr.va_toastrelid);
+		free(toast_relation_path);
 		toast_rel_fp = fopen(toast_relation_filename, "rb");
 		if (!toast_rel_fp) {
 			printf("Cannot open TOAST relation %s\n",
 					toast_relation_filename);
-			result = -1;
+			return -1;
+		}
+
+		toast_relation_block_size = GetBlockSize(toast_rel_fp);
+		toast_data = malloc(toast_ptr.va_rawsize);
+
+		/* Loop until all chunks have been read */
+		while (want_chunk_id < num_chunks)
+		{
+			/* Restart at beginning of file */
+			fseek(toast_rel_fp, 0, SEEK_SET);
+
+			result = DumpFileContents(block_options,
+						control_options,
+						toast_rel_fp,
+						toast_relation_block_size,
+						-1, /* no start block */
+						-1, /* no end block */
+						true, /* is toast relation */
+						toast_ptr.va_valueid,
+						&want_chunk_id,
+						toast_ext_size,
+						toast_data,
+						&toastDataRead);
+
+			if (loops++ > num_chunks)
+			{
+				printf("Not all TOAST chunks found after scanning TOAST table %" PRIu32 " times, giving up\n",
+						num_chunks);
+				result = -1;
+			}
+
+			if (result != 0)
+				break;
+		}
+
+		fclose(toast_rel_fp);
+
+		if (result == 0)
+		{
+			if (VARATT_EXTERNAL_IS_COMPRESSED(toast_ptr))
+				result = DumpCompressedString(toast_data, toast_ext_size, parse_value);
+			else
+				result = parse_value(toast_data, toast_ext_size);
 		}
 		else
 		{
-			unsigned int toast_relation_block_size = GetBlockSize(toast_rel_fp);
-			fseek(toast_rel_fp, 0, SEEK_SET);
-			toast_data = malloc(toast_ptr.va_rawsize);
-
-			result = DumpFileContents(block_options,
-					control_options,
-					toast_rel_fp,
-					toast_relation_block_size,
-					-1, /* no start block */
-					-1, /* no end block */
-					true, /* is toast relation */
-					toast_ptr.va_valueid,
-					toast_ext_size,
-					toast_data);
-
-			if (result == 0)
-			{
-				if (VARATT_EXTERNAL_IS_COMPRESSED(toast_ptr))
-					result = DumpCompressedString(toast_data, toast_ext_size, parse_value);
-				else
-					result = parse_value(toast_data, toast_ext_size);
-			}
-			else
-			{
-				printf("Error in TOAST file.\n");
-			}
-
-			free(toast_data);
-			fclose(toast_rel_fp);
+			printf("Error in TOAST file.\n");
 		}
 
-		free(toast_relation_path);
+		free(toast_data);
+
 	}
 	/* If tag is indirect or expanded, it was stored in memory. */
 	else
@@ -1513,7 +1530,9 @@ void
 ToastChunkDecode(const char *tuple_data,
 		unsigned int tuple_size,
 		Oid toast_oid,
+		Oid *read_toast_oid,
 		uint32 *chunk_id,
+		uint32 *want_chunk_id,
 		char *chunk_data,
 		unsigned int *chunk_data_size)
 {
@@ -1521,14 +1540,13 @@ ToastChunkDecode(const char *tuple_data,
 	const char	   *data = tuple_data + header->t_hoff;
 	unsigned int	size = tuple_size - header->t_hoff;
 	unsigned int	processed_size = 0;
-	Oid				read_toast_oid;
 	int				ret;
 
 	*chunk_data_size = 0;
 	*chunk_id = 0;
 
 	/* decode toast_id */
-	ret = DecodeOidBinary(data, size, &processed_size, &read_toast_oid);
+	ret = DecodeOidBinary(data, size, &processed_size, read_toast_oid);
 	if (ret < 0)
 	{
 		printf("Error: unable to decode a TOAST tuple toast_id, "
@@ -1547,7 +1565,7 @@ ToastChunkDecode(const char *tuple_data,
 	}
 
 	/* It is not what we are looking for */
-	if (toast_oid != read_toast_oid)
+	if (toast_oid != *read_toast_oid)
 		return;
 
 	/* decode chunk_id */
@@ -1560,6 +1578,10 @@ ToastChunkDecode(const char *tuple_data,
 		return;
 	}
 
+	/* Not the chunk we want to read next */
+	if (*chunk_id != *want_chunk_id)
+		return;
+
 	size -= processed_size;
 	data += processed_size;
 	if (size <= 0)
@@ -1586,4 +1608,7 @@ ToastChunkDecode(const char *tuple_data,
 				"Partial data: %s\n", size, copyString.data);
 		return;
 	}
+
+	/* advance to next chunk */
+	*want_chunk_id += 1;
 }
