@@ -3,7 +3,7 @@
  *				   formatting heap (data), index and control files.
  *
  * Copyright (c) 2002-2010 Red Hat, Inc.
- * Copyright (c) 2011-2024, PostgreSQL Global Development Group
+ * Copyright (c) 2011-2025, PostgreSQL Global Development Group
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -22,7 +22,7 @@
  * Original Author: Patrick Macdonald <patrickm@redhat.com>
  */
 
-#define FD_VERSION	"17.1"		/* version ID of pg_filedump */
+#define FD_VERSION	"17.4"		/* version ID of pg_filedump */
 #define FD_PG_VERSION	"PostgreSQL 8.x .. 17.x"		/* PG version it works with */
 
 #include "postgres.h"
@@ -154,4 +154,4 @@ unsigned int GetBlockSize(FILE *fp);
 int DumpFileContents(unsigned int blockOptions, unsigned int controlOptions,
 					 FILE *fp, unsigned int blockSize, int blockStart,
 					int blockEnd, bool isToast, Oid toastOid,
-					unsigned int toastExternalSize, char *toastValue);
+					uint32 *want_chunk_id, unsigned int toastExternalSize, char *toastValue, unsigned int *toastDataRead);
